package geniusweb.protocol.session;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import geniusweb.actions.Action;
import geniusweb.actions.PartyId;
import geniusweb.progress.Progress;
import geniusweb.protocol.ProtocolException;
import geniusweb.protocol.partyconnection.ProtocolToPartyConnections;
import geniusweb.references.PartyWithProfile;

/**
 * The default current state of the session. immutable.
 */
public abstract class DefaultSessionState implements SessionState {
	private final List<Action> actions;

	private final transient ProtocolToPartyConnections connections;

	private final Progress progress;

	private final SessionSettings settings;

	private final ProtocolException error;

	private final Map<PartyId, PartyWithProfile> partyprofiles;

	// only for deserializer.
	protected DefaultSessionState() {
		actions = null;
		connections = null;
		progress = null;
		settings = null;
		partyprofiles = null;
		error = null;
	}

	/**
	 * 
	 * @param actions       value for actions done so far. null equals to empty
	 *                      list
	 * @param conns         the currently existing connections. Can be empty. If
	 *                      null it is assumed to be empty. Each connection
	 *                      represents another party. Normally there is exactly
	 *                      1 connection for every party. The protocol should
	 *                      check this.
	 * @param progr         the {@link Progress} that governs this session. Can
	 *                      be null if session did not yet start.
	 * @param settings      the settings used for the session
	 * @param partyprofiles map with the {@link PartyWithProfile} for connected
	 *                      parties. null is equivalent to an empty map.
	 * @param e             the exception that occured, usually null. All errors
	 *                      occuring due to faulty {@link Action}s translate to
	 *                      {@link ProtocolException}s. All errors in our own
	 *                      code are bugs (not ProtocolExceptions) and should
	 *                      result in a throw and terminate the session.
	 */
	public DefaultSessionState(List<Action> actions,
			ProtocolToPartyConnections conns, Progress progr,
			SessionSettings settings,
			Map<PartyId, PartyWithProfile> partyprofiles, ProtocolException e) {
		if (partyprofiles == null) {
			partyprofiles = Collections.emptyMap();
		}
		if (conns == null) {
			this.connections = new ProtocolToPartyConnections(
					Collections.emptyList());
		} else {
			this.connections = conns;
		}
		if (actions == null) {
			this.actions = new LinkedList<>();
		} else {
			this.actions = actions;
		}

		if (!connections.allunique()) {
			throw new IllegalArgumentException(
					"There can not be multiple connections for a party:"
							+ connections);
		}
		this.progress = progr;
		this.settings = settings;
		this.partyprofiles = partyprofiles;
		this.error = e;
	}

	/**
	 * @return existing connections.
	 */
	public ProtocolToPartyConnections getConnections() {
		return connections;
	}

	/**
	 * @return map with {@link PartyWithProfile} for the parties. May be an
	 *         incomplete map, as more parties with their profiles may be set
	 *         only later.
	 */
	public Map<PartyId, PartyWithProfile> getPartyProfiles() {
		return partyprofiles;
	}

	/**
	 * 
	 * @return unmodifyable list of actions done so far.
	 */
	@Override
	public List<Action> getActions() {
		return Collections.unmodifiableList(actions);
	}

	@Override
	public Progress getProgress() {
		return progress;
	}

	@Override
	public SessionSettings getSettings() {
		return settings;
	}

	@Override
	public boolean isFinal(long currentTimeMs) {
		return error != null
				|| (progress != null && progress.isPastDeadline(currentTimeMs));
	}

	public ProtocolException getError() {
		return error;
	}

	@Override
	public String toString() {
		return this.getClass().getSimpleName() + "[" + actions + ","
				+ connections + "," + progress + "," + settings + "," + error
				+ "]";
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((actions == null) ? 0 : actions.hashCode());
		result = prime * result
				+ ((connections == null) ? 0 : connections.hashCode());
		result = prime * result + ((error == null) ? 0 : error.hashCode());
		result = prime * result
				+ ((partyprofiles == null) ? 0 : partyprofiles.hashCode());
		result = prime * result
				+ ((progress == null) ? 0 : progress.hashCode());
		result = prime * result
				+ ((settings == null) ? 0 : settings.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		DefaultSessionState other = (DefaultSessionState) obj;
		if (actions == null) {
			if (other.actions != null)
				return false;
		} else if (!actions.equals(other.actions))
			return false;
		if (connections == null) {
			if (other.connections != null)
				return false;
		} else if (!connections.equals(other.connections))
			return false;
		if (error == null) {
			if (other.error != null)
				return false;
		} else if (!error.equals(other.error))
			return false;
		if (partyprofiles == null) {
			if (other.partyprofiles != null)
				return false;
		} else if (!partyprofiles.equals(other.partyprofiles))
			return false;
		if (progress == null) {
			if (other.progress != null)
				return false;
		} else if (!progress.equals(other.progress))
			return false;
		if (settings == null) {
			if (other.settings != null)
				return false;
		} else if (!settings.equals(other.settings))
			return false;
		return true;
	}

}
